package cn.superdesk.libs.mybatis.plugin.pagination;

import java.lang.reflect.Method;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.*;

import cn.superdesk.libs.common.json.JsonUtils;
import cn.superdesk.libs.mybatis.exception.MybatisHanlerInitException;
import cn.superdesk.libs.mybatis.plugin.pagination.annotation.Pageable;
import org.apache.commons.lang3.StringUtils;
import org.apache.ibatis.cache.CacheKey;
import org.apache.ibatis.executor.Executor;
import org.apache.ibatis.mapping.*;
import org.apache.ibatis.plugin.Invocation;
import org.apache.ibatis.session.Configuration;
import org.apache.ibatis.session.ResultHandler;
import org.apache.ibatis.session.RowBounds;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import cn.superdesk.libs.mybatis.MybatisConfigs;
import cn.superdesk.libs.mybatis.core.InterceptorHandler;
import cn.superdesk.libs.mybatis.parser.EntityInfo;
import cn.superdesk.libs.mybatis.parser.MybatisMapperParser;
import cn.superdesk.libs.mybatis.plugin.SuperdeskMybatisInterceptor;
import cn.superdesk.libs.mybatis.plugin.pagination.PageSqlUtils.DbType;

public class PaginationHandler implements InterceptorHandler {

	private static Logger logger = LoggerFactory.getLogger(PaginationHandler.class);

	public static final  String NAME              = "page";
	private static final String PAGE_COUNT_SUFFIX = "_PageCount";

	private Map<String, Boolean> pageMappedStatements = new HashMap<>();

	private DbType dbType = DbType.MYSQL;

	/**
	 * @param dbType
	 */
	public void setDbType(String dbType) {

		if (StringUtils.isBlank(dbType)) {
			return;
		}

		DbType[] dbTypes = DbType.values();

		for (DbType dt : dbTypes) {
			if (dt.name().equalsIgnoreCase(dbType)) {
				this.dbType = dt;
				break;
			}
		}
	}

	@Override
	public void start(SuperdeskMybatisInterceptor context) {

		setDbType(MybatisConfigs.getDbType(context.getGroupName()));

		logger.debug("dbType:{}", dbType.name());

		List<EntityInfo> entityInfos = MybatisMapperParser.getEntityInfos(context.getGroupName());

		for (EntityInfo ei : entityInfos) {

			logger.debug("PaginationHandler.start: \nEntityInfo:{}\n", JsonUtils.toPrettyJson(ei));

			Class<?> mapperClass = ei.getMapperClass();

			Method[] methods = mapperClass.getDeclaredMethods();
			for (Method method : methods) {
				if (method.getReturnType() == Page.class) {
					String msId = ei.getMapperClass().getName() + "." + method.getName();

					boolean    withPageParams = false;
					Class<?>[] parameterTypes = method.getParameterTypes();
					self:
					for (Class<?> clazz : parameterTypes) {
						if (withPageParams = (clazz == PageParams.class || clazz.getSuperclass() == PageParams.class)) {
							break self;
						}
					}

					if (!withPageParams) {
						throw new MybatisHanlerInitException(String.format("method[%s] returnType is:Page,but not found Parameter[PageParams] in Parameters list", method.getName()));
					}
					pageMappedStatements.put(msId, true);

				} else if (method.isAnnotationPresent(Pageable.class)) {

					String msId = ei.getMapperClass().getName() + "." + method.getName();
					pageMappedStatements.put(msId, false);

				}
			}
		}
	}

	/**
	 * @param invocation
	 * @return
	 * @throws Throwable
	 */
	@SuppressWarnings({"rawtypes", "unchecked"})
	@Override
	public Object onInterceptor(Invocation invocation) throws Throwable {
		try {

			// 原来的代码
			final Executor executor = (Executor) invocation.getTarget();
			final Object[] args     = invocation.getArgs();
			final MappedStatement mappedStatement = (MappedStatement) args[0];

			if (!mappedStatement.getSqlCommandType().equals(SqlCommandType.SELECT)) {
				return null;
			}

			PageParams pageParams = PageExecutor.getPageParams();
			if (pageParams == null && !pageMappedStatements.keySet().contains(mappedStatement.getId())) {
				return null;
			}


//			Connection          connection    = (Connection) args[0];
			final Object        parameter     = args[1];
			final RowBounds     rowBounds     = (RowBounds) args[2];
			final ResultHandler resultHandler = (ResultHandler) args[3];

			BoundSql boundSql;
			if (args.length == 4) {
				boundSql = mappedStatement.getBoundSql(parameter);
			} else {
				boundSql = (BoundSql) args[5];
			}

			// 处理分页 参数 ============================== start ==============================
			if (pageParams == null && pageMappedStatements.get(mappedStatement.getId())) {
				if (parameter instanceof Map) {
					Collection parameterValues = ((Map) parameter).values();
					for (Object val : parameterValues) {
						if (val instanceof PageParams) {
							pageParams = (PageParams) val;
							break;
						}
					}
				} else {
					pageParams = (PageParams) parameter;
				}
			}

			/*
			 * 不需要分页的场合，如果 size 小于 0 返回结果集
			 */
			if (pageParams == null || pageParams.getPageSize() < 0) {
				return null;
			}

			/*
			 * 处理单页条数限制
			 */
//			if (limit > 0 && limit <= pageParams.getPageSize()) {
//				pageParams.setPageSize(limit);
//			}
			// 处理分页 参数 ==============================  end  ==============================

			// 查询总数
			MappedStatement countMappedStatement = getCountMappedStatement(mappedStatement);
			Long            total                = executeQueryCount(executor, countMappedStatement, parameter, boundSql, rowBounds, resultHandler);

			// 查询分页数据
			List<?> datas = executeQuery(executor, mappedStatement, parameter, boundSql, rowBounds, resultHandler, pageParams);


			// 整合 返回
			Page<Object>  page = new Page<Object>(pageParams, total, (List<Object>) datas);
			List<Page<?>> list = new ArrayList<Page<?>>(1);
			list.add(page);
			return list;

		} finally {
			PageExecutor.clearPageParams();
		}
	}


	@SuppressWarnings("rawtypes")
	private Long executeQueryCount(Executor executor,
								   MappedStatement countMs,
								   Object parameter,
								   BoundSql boundSql,
								   RowBounds rowBounds,
								   ResultHandler resultHandler
	) throws IllegalAccessException, SQLException {

		CacheKey countKey = executor.createCacheKey(countMs, parameter, RowBounds.DEFAULT, boundSql);


		logger.debug("executeQueryCount.boundSql.getSql() : \n" + boundSql.getSql());

		String orignSql = StringUtils.replace(boundSql.getSql(), ";$", StringUtils.EMPTY);
		logger.debug("executeQueryCount.orignSql : \n" + orignSql);

		// count sql 就是TMD这个地方用 正则 分割的
		String countSql = PageSqlUtils.getCountSql(orignSql);
		logger.debug("executeQueryCount.countSql : \n" + countSql);

		BoundSql countBoundSql = new BoundSql(
				countMs.getConfiguration(),
				countSql,
				boundSql.getParameterMappings(),
				parameter);

//		for (ParameterMapping parameterMapping : boundSql.getParameterMappings()) {
//			String propertyName = parameterMapping.getProperty();
//			if(boundSql.hasAdditionalParameter(propertyName)){
//				countBoundSql.setAdditionalParameter(propertyName, boundSql.getAdditionalParameter(propertyName));
//			}
//		}

		// 执行 count 查询
		Object countResultList = executor.query(
				countMs,
				parameter,
				RowBounds.DEFAULT,
				resultHandler,
				countKey,
				countBoundSql);

		Long count = (Long) ((List) countResultList).get(0);

		return count;
	}

	@SuppressWarnings("rawtypes")
	private List executeQuery(Executor executor,
							  MappedStatement ms,
							  Object parameter,
							  BoundSql boundSql,
							  RowBounds rowBounds,
							  ResultHandler resultHandler,
							  PageParams pageParams
	) throws IllegalAccessException, SQLException {

		CacheKey countKey = executor.createCacheKey(ms, parameter, RowBounds.DEFAULT, boundSql);

		String orignSql = StringUtils.replace(boundSql.getSql(), ";$", StringUtils.EMPTY);

		String pageSql = PageSqlUtils.getLimitSQL(dbType, orignSql, pageParams);

		BoundSql countBoundSql = new BoundSql(
				ms.getConfiguration(),
				pageSql,
				boundSql.getParameterMappings(),
				parameter);

		List<?> resultList = executor.query(
				ms,
				parameter,
				RowBounds.DEFAULT,
				resultHandler,
				countKey,
				countBoundSql);

		return resultList;
	}

	/**
	 * 新建count查询的MappedStatement
	 *
	 * @param ms
	 * @return
	 */
	public MappedStatement getCountMappedStatement(MappedStatement ms) {

		String newMsId = ms.getId() + PAGE_COUNT_SUFFIX;

		MappedStatement statement     = null;
		Configuration   configuration = ms.getConfiguration();

		try {
			statement = configuration.getMappedStatement(newMsId);
			if (statement != null) return statement;
		} catch (Exception e) {
		}

		synchronized (configuration) {

			if (configuration.hasStatement(newMsId)) return configuration.getMappedStatement(newMsId);

			MappedStatement.Builder builder = new MappedStatement.Builder(ms.getConfiguration(), newMsId, ms.getSqlSource(), ms.getSqlCommandType());
			builder.resource(ms.getResource());
			builder.fetchSize(ms.getFetchSize());
			builder.statementType(ms.getStatementType());
			builder.keyGenerator(ms.getKeyGenerator());

			if (ms.getKeyProperties() != null && ms.getKeyProperties().length != 0) {
				StringBuilder keyProperties = new StringBuilder();
				for (String keyProperty : ms.getKeyProperties()) {
					keyProperties.append(keyProperty).append(",");
				}
				keyProperties.delete(keyProperties.length() - 1, keyProperties.length());
				builder.keyProperty(keyProperties.toString());
			}

			builder.timeout(ms.getTimeout());
			builder.parameterMap(ms.getParameterMap());

			//count查询返回值int
			List<ResultMap> resultMaps = new ArrayList<ResultMap>();
			String          id         = newMsId + "-Inline";
			ResultMap       resultMap  = new ResultMap.Builder(configuration, id, Long.class, new ArrayList<ResultMapping>(0)).build();
			resultMaps.add(resultMap);
			builder.resultMaps(resultMaps);

			builder.resultSetType(ms.getResultSetType());
			builder.cache(ms.getCache());
			builder.flushCacheRequired(ms.isFlushCacheRequired());
			builder.useCache(ms.isUseCache());

			statement = builder.build();
			configuration.addMappedStatement(statement);

			return statement;
		}

	}

	@Override
	public void onFinished(Invocation invocation, Object result) {

	}

	@Override
	public void close() {
	}

	@Override
	public int interceptorOrder() {
		return 9;
	}

}
