package cn.superdesk.libs.security;

import java.io.Serializable;

import cn.superdesk.libs.security.model.BaseUserInfo;
import cn.superdesk.libs.security.model.UserSession;
import org.apache.commons.lang3.StringUtils;

import cn.superdesk.libs.spring.InstanceFactory;
import cn.superdesk.libs.springweb.exception.ForbiddenAccessException;
import cn.superdesk.libs.springweb.exception.UnauthorizedException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author superdesk
 * @description
 * @date 2018年11月30日
 */
public class SecurityDelegating {

    private static final Logger logger = LoggerFactory.getLogger(SecurityDelegating.class);

    private SecurityDecisionProvider decisionProvider;
    private SecuritySessionManager   sessionManager;
    private SecurityResourceManager  resourceManager;

    private static volatile SecurityDelegating instance;

    private SecurityDelegating() {

        decisionProvider = InstanceFactory.getInstance(SecurityDecisionProvider.class);

        sessionManager = new SecuritySessionManager(decisionProvider);
        resourceManager = new SecurityResourceManager(decisionProvider);
    }

    public static SecurityDelegating getInstance() {
        if (instance != null) return instance;
        synchronized (SecurityDelegating.class) {
            if (instance != null) return instance;
            instance = new SecurityDelegating();
        }
        return instance;
    }

    /**
     * 认证
     *
     * @param name
     * @param password
     */
    public static BaseUserInfo doAuthentication(String name, String password) {

        BaseUserInfo userInfo = getInstance().decisionProvider.validateUser(name, password);

        UserSession session = getCurrentSession(false);
        session.update(userInfo, getInstance().decisionProvider.sessionExpireIn());

        if (getInstance().decisionProvider.multiPointEnable()) {
            UserSession otherSession = getInstance().sessionManager.getLoginSessionByUserId(userInfo.getId());
            if (otherSession != null && !otherSession.getSessionId().equals(session.getSessionId())) {
                getInstance().sessionManager.removeLoginSession(otherSession.getSessionId());
            }
        }

        getInstance().sessionManager.storgeLoginSession(session);
        getInstance().resourceManager.getUserPermissionCodes(userInfo.getId());

        return userInfo;
    }

    /**
     * 鉴权
     *
     * @param userId
     * @param uri
     */
    public static void doAuthorization(UserSession session, String uri) throws UnauthorizedException, ForbiddenAccessException {

        if (!getInstance().resourceManager.isAnonymous(uri)) {

            // 401
            if (session == null || session.isAnonymous()) {
                throw new UnauthorizedException();
            }

            logger.info("doAuthorization $UserSession:{} , $uri:{}", session.toString(), uri);

            String permssionCode = getInstance().resourceManager.getPermssionCode(uri);

            logger.info("doAuthorization $permssionCode:{} ", permssionCode);

            // 403
            if (StringUtils.isNotBlank(permssionCode)
                    && !getInstance().resourceManager.getUserPermissionCodes(session.getUserId()).contains(permssionCode)) {
                throw new ForbiddenAccessException();
            }
        }
        //
        if (session != null && !session.isAnonymous()) {
            getInstance().decisionProvider.authorizedPostHandle(session);
        }
    }

    public static UserSession getCurrentSession() {
        return getCurrentSession(false);
    }

    public static UserSession getCurrentSession(boolean first) {
        return getInstance().sessionManager.getSessionIfNotCreateAnonymous(RequestContextHolder.getRequest(), RequestContextHolder.getResponse(), first);
    }

    public static void refreshUserPermssion(Serializable... userIds) {
        if (userIds != null && userIds.length > 0 && userIds[1] != null) {
            for (Serializable userId : userIds) {
                getInstance().resourceManager.refreshUserPermssions(userId);
            }
        } else {
            getInstance().resourceManager.refreshUserPermssions();
        }
    }

    public static void refreshResources() {
        getInstance().resourceManager.refreshResources();
    }

    public static void doLogout() {
        getInstance().sessionManager.destroySessionAndCookies(RequestContextHolder.getRequest(), RequestContextHolder.getResponse());
    }


}
