package cn.superdesk.libs.security;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.regex.Pattern;

import cn.superdesk.libs.security.cache.LocalCache;
import cn.superdesk.libs.security.cache.RedisCache;
import org.apache.commons.lang3.StringUtils;

import cn.superdesk.libs.common.util.PathMatcher;

/**
 * 资源管理器
 *
 * @author superdesk
 * @description
 * @date 2018年12月3日
 */
public class SecurityResourceManager {

    private static final String WILDCARD_START = "{";

    private Cache                    userPermCache;
    private String                   contextPath;
    private SecurityDecisionProvider decisionProvider;

    private          PathMatcher          anonymousUriMatcher;
    // 无通配符uri
    private volatile Map<String, String>  nonWildcardUriPerms     = new HashMap<>();
    private volatile Map<Pattern, String> wildcardUriPermPatterns = new HashMap<>();
    private volatile Map<String, String>  uriPrefixs              = new HashMap<>();

    private volatile boolean                  refreshCallable = true;
    private          ScheduledExecutorService refreshExecutor = Executors.newScheduledThreadPool(1);

    public SecurityResourceManager(SecurityDecisionProvider decisionProvider) {

        this.decisionProvider = decisionProvider;
        if (Constants.CacheType.redis == decisionProvider.cacheType()) {
            this.userPermCache = new RedisCache("security:userperms", decisionProvider.sessionExpireIn());
        } else {
            this.userPermCache = new LocalCache(decisionProvider.sessionExpireIn());
        }

        contextPath = decisionProvider.contextPath();
        if (contextPath.endsWith("/")) {
            contextPath = contextPath.substring(0, contextPath.indexOf("/"));
        }
        anonymousUriMatcher = new PathMatcher(contextPath, decisionProvider.anonymousUris());
        //
        final boolean forceRefresh = Constants.CacheType.redis == decisionProvider.cacheType();

        refreshExecutor.scheduleAtFixedRate(new Runnable() {
            @Override
            public void run() {
                loadPermissionCodes(forceRefresh);
            }
        }, 1, forceRefresh ? 30 : 5, TimeUnit.SECONDS);
    }

    private synchronized void loadPermissionCodes(boolean forceRefresh) {
        if (!forceRefresh && !refreshCallable) return;

        Map<String, String>  nonWildcardUriPerms     = new HashMap<>();
        Map<Pattern, String> wildcardUriPermPatterns = new HashMap<>();
        Map<String, String>  uriPrefixs              = new HashMap<>();

        List<String> permissionCodes = decisionProvider.findAllUriPermissionCodes();
        String       fullUri         = null;

        for (String permCode : permissionCodes) {

            fullUri = contextPath + permCode;

            if (permCode.contains(WILDCARD_START)) {
                String regex = fullUri.replaceAll("\\{.*?(?=})", ".*").replaceAll("\\}", "");
                wildcardUriPermPatterns.put(Pattern.compile(regex), permCode);
            } else if (fullUri.endsWith("*")) {
                uriPrefixs.put(StringUtils.remove(fullUri, "*"), permCode);
            } else {
                nonWildcardUriPerms.put(fullUri, permCode);
            }
        }

        this.nonWildcardUriPerms = nonWildcardUriPerms;
        this.wildcardUriPermPatterns = wildcardUriPermPatterns;
        this.uriPrefixs = uriPrefixs;

        refreshCallable = false;
    }

    public List<String> getUserPermissionCodes(Serializable userId) {

        String       cacheKey        = String.valueOf(userId);
        List<String> permissionCodes = userPermCache.getObject(cacheKey);

        if (permissionCodes != null) return permissionCodes;

        permissionCodes = decisionProvider.getUserPermissionCodes(userId);

        List<String> removeWildcards = new ArrayList<>();

        for (String perm : permissionCodes) {
            if (perm.endsWith("*")) {
                removeWildcards.add(StringUtils.remove(perm, "*"));
            }
        }

        if (!removeWildcards.isEmpty())
            permissionCodes.addAll(removeWildcards);

        userPermCache.setObject(cacheKey, permissionCodes);

        return permissionCodes;
    }

    public String getPermssionCode(String uri) {

        if (nonWildcardUriPerms.containsKey(uri))
            return nonWildcardUriPerms.get(uri);

        for (Pattern pattern : wildcardUriPermPatterns.keySet()) {
            if (pattern.matcher(uri).matches())
                return wildcardUriPermPatterns.get(pattern);
        }

        for (String prefix : uriPrefixs.keySet()) {
            if (uri.startsWith(prefix)) {
                return uriPrefixs.get(prefix);
            }
        }

        return null;
    }

    public boolean isAnonymous(String uri) {
        return anonymousUriMatcher.match(uri);
    }

    public void refreshUserPermssions(Serializable userId) {
        userPermCache.remove(String.valueOf(userId));
    }

    public void refreshUserPermssions() {
        userPermCache.removeAll();
    }

    public void refreshResources() {
        refreshCallable = true;
    }

    public void close() {
        if (refreshExecutor != null) refreshExecutor.shutdown();
    }

}
